"""
Block Neural Autoregressive bijection implementation.
"""

import math
from typing import Callable, Optional, Union, Tuple

import jax
import jax.numpy as jnp
from jax import random
from jax.random import KeyArray

from flowjax.bijections import Bijection
from flowjax.nn.block_autoregressive import BlockAutoregressiveLinear, BlockTanh


class BlockAutoregressiveNetwork(Bijection):
    """Block Autoregressive Network (https://arxiv.org/abs/1904.04676)."""

    depth: int
    layers: list
    block_dim: int
    activation: Callable

    def __init__(
        self,
        key: KeyArray,
        dim: int,
        cond_dim: Union[None, int],
        depth: int,
        block_dim: int,
        activation: Optional[Callable] = None,
    ):
        """
        Args:
            key (KeyArray): Jax PRNGKey
            dim (int): Dimension of the distribution.
            cond_dim (Union[None, Tuple[int]]): Dimension of extra conditioning variables.
            depth (int): Number of hidden layers in the network.
            block_dim (int): Block dimension (hidden layer size is `dim*block_dim`).
            activation (Callable, optional): Activation function. Defaults to BlockTanh.
        """
        activation = BlockTanh(dim) if activation is None else activation
        layers = []
        if depth == 0:
            layers.append(BlockAutoregressiveLinear(key, dim, (1, 1), cond_dim))
        else:
            keys = random.split(key, depth + 1)

            block_shapes = [
                (block_dim, 1),
                *[(block_dim, block_dim)] * (depth - 1),
                (1, block_dim),
            ]
            cond_dims = [cond_dim] + [None] * depth

            for key, block_shape, cd in zip(keys, block_shapes, cond_dims):

                layers.extend(
                    [
                        BlockAutoregressiveLinear(key, dim, block_shape, cd),
                        activation,
                    ]
                )
            layers = layers[:-1]  # remove last activation

        self.depth = depth
        self.layers = layers
        self.block_dim = block_dim
        self.activation = activation
        self.shape = (dim,)
        self.cond_shape = (cond_dim,) if cond_dim is not None else None

    def transform(self, x, condition=None):
        self._argcheck(x, condition)
        x = self.layers[0](x, condition)[0]
        for layer in self.layers[1:]:
            x = layer(x)[0]
        return x

    def transform_and_log_abs_det_jacobian(self, x, condition=None):
        self._argcheck(x, condition)
        x, log_det_3d_0 = self.layers[0](x, condition)
        log_det_3ds = [log_det_3d_0]
        for layer in self.layers[1:]:
            x, log_det_3d = layer(x)
            log_det_3ds.append(log_det_3d)

        logdet = log_det_3ds[-1]
        for ld in reversed(log_det_3ds[:-1]):  # TODO avoid loop
            logdet = logmatmulexp(logdet, ld)
        return x, logdet.sum()

    def inverse(self, *args, **kwargs):
        raise NotImplementedError(
            "This transform would require numerical methods for inversion."
        )

    def inverse_and_log_abs_det_jacobian(self, *args, **kwargs):
        raise NotImplementedError(
            "This transform would require numerical methods for inversion."
        )


def logmatmulexp(x, y):
    """
    Numerically stable version of ``(x.log() @ y.log()).exp()``. From numpyro https://github.com/pyro-ppl/numpyro/blob/f2ff89a3a7147617e185eb51148eb15d56d44661/numpyro/distributions/util.py#L387
    """
    x_shift = jax.lax.stop_gradient(jnp.amax(x, -1, keepdims=True))
    y_shift = jax.lax.stop_gradient(jnp.amax(y, -2, keepdims=True))
    xy = jnp.log(jnp.matmul(jnp.exp(x - x_shift), jnp.exp(y - y_shift)))
    return xy + x_shift + y_shift



from functools import partial
import equinox as eqx
from flowjax.bijections import Bijection
from jax.lax import scan
from typing import Any, Tuple


class Vmap(Bijection):
    """Expand the dimension of a bijection by vmapping. By default, we vmap over both the
    bijection parameters and x, although this behaviour can be modified by providing key
    word arguments that are passed to ``equinox.filter_vmap``. The arguments names for
    the vmapped functions are (bijection, x).
    
    Vmapping over the conditioning variable is not currently supported.

    Example:
        Affine parameters usually act elementwise, but we could vmap excluding the
        the bijection to create a global affine (sharing the location and scale).
        
        .. doctest::

            >>> from flowjax.bijections import Vmap, Affine
            >>> import jax.numpy as jnp
            >>> affine = Vmap(Affine(1), (3,), kwargs=dict(bijection=None))
            >>> affine.transform(jnp.ones(3))
            Array([2., 2., 2.], dtype=float32)

    """

    ndim_to_add: int
    bijection: Bijection
    kwargs: dict

    def __init__(self, bijection: Bijection, shape: Tuple[int], **kwargs):
        """
        

        Args:
            bijection (Bijection): Bijection. If vmapping over the bijection, the array leaves
                in bijection should have additional leading axes with shape equalling `shape`.
                Often it is convenient to construct these using `equinox.filter_vmap`.
            shape (Tuple[int]): Shape prepended to the bijection shape. If len(shape)>1, multiple applications of vmap will be used.
            **kwargs: kwargs, passed to equinox.filter_vmap, allowing e.g. control over which variables to map over.
        """

        self.bijection = bijection
        self.ndim_to_add = len(shape)
        self.kwargs = kwargs  # For filter vmap
        self.shape = (
            shape + self.bijection.shape if self.bijection.shape is not None else None
        )
        self.cond_shape = bijection.cond_shape

    def transform(self, x, condition=None):
        self._argcheck(x, condition)
        f = lambda bijection, x: bijection.transform(x, condition)
        f = self._multivmap(f)
        return f(self.bijection, x)

    def inverse(self, y, condition=None):
        self._argcheck(y, condition)
        f = lambda bijection, x: bijection.inverse(x, condition)
        f = self._multivmap(f)
        return f(self.bijection, y)

    def transform_and_log_abs_det_jacobian(self, x, condition=None):
        self._argcheck(x, condition)
        f = lambda bijection, x: bijection.transform_and_log_abs_det_jacobian(x, condition)
        f = self._multivmap(f)
        y, log_det = f(self.bijection, x)
        return y, log_det.sum()

    def inverse_and_log_abs_det_jacobian(self, y, condition=None):
        self._argcheck(y, condition)
        f = lambda bijection, x: bijection.inverse_and_log_abs_det_jacobian(x, condition)
        f = self._multivmap(f)
        x, log_det = f(self.bijection, y)
        return x, log_det.sum()

    def _multivmap(self, f):
        "Compose Vmap to add ndim batch dimensions."
        for _ in range(self.ndim_to_add):
            f = eqx.filter_vmap(f, **self.kwargs)
        return f


class Scan(Bijection):
    """Repeatedly apply the same bijection with different parameter values. Internally,
    uses `jax.lax.scan` to reduce compilation time."""

    static: Any
    params: Any

    def __init__(self, bijection: Bijection):
        """
        The array leaves in `bijection` should have an additional leading axis to scan over.
        Often it is convenient to construct these using `equinox.filter_vmap`.

        Args:
            bijection (Bijection): A bijection, in which the arrays leaves have an additional leading axis to scan over.
                For complex bijections, it can be convenient to create compatible bijections with ``equinox.filter_vmap``.

        Example:
            Below is equivilent to ``Chain([Affine(p) for p in params])``.

            .. doctest::

                >>> from flowjax.bijections import Scan, Affine
                >>> import jax.numpy as jnp
                >>> import equinox as eqx
                >>> params = jnp.ones((3, 2))
                >>> affine = Scan(Affine(params))

        """
        self.params, self.static = eqx.partition(bijection, eqx.is_array)  # type: ignore
        self.shape = bijection.shape
        self.cond_shape = bijection.cond_shape

    def transform(self, x, condition=None):
        self._argcheck(x, condition)

        def fn(x, p, condition=None):
            bijection = eqx.combine(self.static, p)
            result = bijection.transform(x, condition)  # type: ignore
            return (result, None)

        fn = partial(fn, condition=condition)
        y, _ = scan(fn, x, self.params)
        return y

    def transform_and_log_abs_det_jacobian(self, x, condition=None):
        self._argcheck(x, condition)

        def fn(carry, p, condition):
            x, log_det = carry
            bijection = eqx.combine(self.static, p)
            y, log_det_i = bijection.transform_and_log_abs_det_jacobian(x, condition)  # type: ignore
            return ((y, log_det + log_det_i.sum()), None)

        fn = partial(fn, condition=condition)
        (y, log_det), _ = scan(fn, (x, 0), self.params)
        return y, log_det

    def inverse(self, y, condition=None):
        self._argcheck(y, condition)

        def fn(y, p, condition=None):
            bijection = eqx.combine(self.static, p)
            x = bijection.inverse(y, condition)  # type: ignore
            return (x, None)

        fn = partial(fn, condition=condition)
        x, _ = scan(fn, y, self.params, reverse=True)
        return x

    def inverse_and_log_abs_det_jacobian(self, y, condition=None):
        self._argcheck(y, condition)

        def fn(carry, p, condition=None):
            y, log_det = carry
            bijection = eqx.combine(self.static, p)
            x, log_det_i = bijection.inverse_and_log_abs_det_jacobian(y, condition)  # type: ignore
            return ((x, log_det + log_det_i.sum()), None)

        fn = partial(fn, condition=condition)
        (y, log_det), _ = scan(fn, (y, 0), self.params, reverse=True)
        return y, log_det



from typing import Callable, List

import jax.nn as jnn
import jax.numpy as jnp
from equinox import Module
from equinox.nn import Linear
from jax import random
from jax.random import KeyArray

from flowjax.masks import rank_based_mask
from flowjax.utils import Array, _identity


class MaskedLinear(Module):
    linear: Linear
    mask: Array

    def __init__(self, mask: Array, use_bias: bool = True, *, key: KeyArray):
        """
        Masked linear layer.

        Args:
            mask (Array): Mask with shape (out_features, in_features).
            key (KeyArray): Jax PRNGKey
            use_bias (bool, optional): Whether to include bias terms. Defaults to True.
        """
        self.linear = Linear(mask.shape[1], mask.shape[0], use_bias, key=key)
        self.mask = mask

    def __call__(self, x: Array):
        x = self.linear.weight * self.mask @ x
        if self.linear.bias is not None:
            x = x + self.linear.bias
        return x


class AutoregressiveMLP(Module):
    in_size: int
    out_size: int
    width_size: int
    depth: int
    in_ranks: Array
    out_ranks: Array
    hidden_ranks: Array
    layers: List[MaskedLinear]
    activation: Callable
    final_activation: Callable

    def __init__(
        self,
        in_ranks: Array,
        hidden_ranks: Array,
        out_ranks: Array,
        depth: int,
        activation: Callable = jnn.relu,
        final_activation: Callable = _identity,
        *,
        key
    ) -> None:
        """An autoregressive multilayer perceptron, similar to ``equinox.nn.composed.MLP``.
        Connections will only exist where in_ranks < out_ranks.

        Args:
            in_ranks (Array): Ranks of the inputs.
            hidden_ranks (Array): Ranks of the hidden layer(s).
            out_ranks (Array): Ranks of the outputs.
            depth (int): Number of hidden layers.
            activation (Callable, optional): Activation function. Defaults to jnn.relu.
            final_activation (Callable, optional): Final activation function. Defaults to _identity.
            key (KeyArray): Jax PRNGKey
        """

        masks = []
        if depth == 0:
            masks.append(rank_based_mask(in_ranks, out_ranks, eq=False))
        else:
            masks.append(rank_based_mask(in_ranks, hidden_ranks, eq=True))
            for _ in range(depth - 1):
                masks.append(rank_based_mask(hidden_ranks, hidden_ranks, eq=True))
            masks.append(rank_based_mask(hidden_ranks, out_ranks, eq=False))

        keys = random.split(key, len(masks))
        layers = [MaskedLinear(mask, key=key) for mask, key in zip(masks, keys)]

        self.layers = layers
        self.in_size = len(in_ranks)
        self.out_size = len(out_ranks)
        self.width_size = len(hidden_ranks)
        self.depth = depth
        self.in_ranks = in_ranks
        self.hidden_ranks = hidden_ranks
        self.out_ranks = out_ranks
        self.activation = activation
        self.final_activation = final_activation

    def __call__(self, x: Array):
        """Forward pass.
        Args:
            x: A JAX array with shape (in_size,).
        """
        for layer in self.layers[:-1]:
            x = layer(x)
            x = self.activation(x)
        x = self.layers[-1](x)
        x = self.final_activation(x)
        return x



from typing import Callable, Union
import math
import equinox as eqx
import jax.nn as jnn
import jax.numpy as jnp
from jax.random import KeyArray
from flowjax.bijections import Bijection
from flowjax.utils import get_ravelled_bijection_constructor, Array
from flowjax.bijections.jax_transforms import Vmap


class Coupling(Bijection):
    d: int
    D: int
    transformer_constructor: Callable
    conditioner: eqx.nn.MLP

    def __init__(
        self,
        key: KeyArray,
        transformer: Bijection,
        d: int,
        D: int,
        cond_dim: Union[None, int],
        nn_width: int,
        nn_depth: int,
        nn_activation: Callable = jnn.relu,
    ):
        """Coupling layer implementation (https://arxiv.org/abs/1605.08803).
        Args:
            key (KeyArray): Jax PRNGKey
            transformer (Bijection): Bijection with shape () to be parameterised by the conditioner neural netork.
            d (int): Number of untransformed conditioning variables.
            D (int): Total dimension.
            cond_dim (Union[None, int]): Dimension of additional conditioning variables.
            nn_width (int): Neural network hidden layer width.
            nn_depth (int): Neural network hidden layer size.
            nn_activation (Callable, optional): Neural network activation function. Defaults to jnn.relu.
        """
        if transformer.shape != () or transformer.cond_shape is not None:
            raise ValueError(
                "Currently, only unconditional transformers with shape () are supported."
            )

        constructor, transformer_init_params = get_ravelled_bijection_constructor(
            transformer
        )

        self.transformer_constructor = constructor
        self.d = d
        self.D = D
        self.shape = (D,)
        self.cond_shape = (cond_dim,) if cond_dim is not None else None

        conditioner_output_size = transformer_init_params.size * (D - d)

        conditioner = eqx.nn.MLP(
            in_size=d if cond_dim is None else d + cond_dim,
            out_size=conditioner_output_size,
            width_size=nn_width,
            depth=nn_depth,
            activation=nn_activation,
            key=key,
        )

        # Initialise bias terms to match the provided transformer parameters
        self.conditioner = eqx.tree_at(
            where=lambda mlp: mlp.layers[-1].bias,
            pytree=conditioner,
            replace=jnp.tile(transformer_init_params, D - d),
        )

    def transform(self, x, condition=None):
        x_cond, x_trans = x[: self.d], x[self.d :]
        nn_input = x_cond if condition is None else jnp.hstack((x_cond, condition))
        transformer_params = self.conditioner(nn_input)
        transformer = self._flat_params_to_transformer(transformer_params)
        y_trans = transformer.transform(x_trans)
        y = jnp.hstack((x_cond, y_trans))
        return y

    def transform_and_log_abs_det_jacobian(self, x, condition=None):
        x_cond, x_trans = x[: self.d], x[self.d :]
        nn_input = x_cond if condition is None else jnp.hstack((x_cond, condition))
        transformer_params = self.conditioner(nn_input)
        transformer = self._flat_params_to_transformer(transformer_params)
        y_trans, log_det = transformer.transform_and_log_abs_det_jacobian(x_trans)
        y = jnp.hstack((x_cond, y_trans))
        return y, log_det

    def inverse(self, y, condition=None):
        x_cond, y_trans = y[: self.d], y[self.d :]
        nn_input = x_cond if condition is None else jnp.concatenate((x_cond, condition))
        transformer_params = self.conditioner(nn_input)
        transformer = self._flat_params_to_transformer(transformer_params)
        x_trans = transformer.inverse(y_trans)
        x = jnp.hstack((x_cond, x_trans))
        return x

    def inverse_and_log_abs_det_jacobian(self, y, condition=None):
        x_cond, y_trans = y[: self.d], y[self.d :]
        nn_input = x_cond if condition is None else jnp.concatenate((x_cond, condition))
        transformer_params = self.conditioner(nn_input)
        transformer = self._flat_params_to_transformer(transformer_params)
        x_trans, log_det = transformer.inverse_and_log_abs_det_jacobian(y_trans)
        x = jnp.hstack((x_cond, x_trans))
        return x, log_det

    def _flat_params_to_transformer(
        self, params: Array
    ):  # TODO code repetition with MAF
        "Reshape to dim X params_per_dim, then vmap."
        dim = self.D - self.d
        transformer_params = jnp.reshape(params, (dim, -1))
        transformer = eqx.filter_vmap(self.transformer_constructor)(transformer_params)
        return Vmap(transformer, (dim,))



"""
Premade versions of common flow architetctures from ``flowjax.flows``.
"""
# Note that here although we could chain arbitrary bijections using `Chain`, here,
# we generally opt to use `Scan`, which avoids excessive compilation
# when the flow layers share the same structure.

from typing import Callable, Optional

import equinox as eqx
import jax.nn as jnn
import jax.numpy as jnp
from jax import random
from jax.nn.initializers import glorot_uniform
from jax.random import KeyArray

from flowjax.bijections import (
    AdditiveLinearCondition,
    BlockAutoregressiveNetwork,
    Chain,
    Coupling,
    Flip,
    Invert,
    MaskedAutoregressive,
    Permute,
    Scan,
    TanhLinearTails,
    TriangularAffine,
    RationalQuadraticSpline,
    Bijection,
)
from flowjax.distributions import Distribution, Transformed


class CouplingFlow(
    Transformed
):  # TODO allow flows to work on higher dimensional inputs?
    flow_layers: int
    nn_width: int
    nn_depth: int
    permute_strategy: str

    def __init__(
        self,
        key: KeyArray,
        base_dist: Distribution,
        transformer: Bijection,
        cond_dim: int = None,
        flow_layers: int = 8,
        nn_width: int = 40,
        nn_depth: int = 2,
        nn_activation: Callable = jnn.relu,
        invert: bool = True,
    ):
        """Coupling flow (https://arxiv.org/abs/1605.08803).

        Args:
            key (KeyArray): Jax PRNGKey.
            base_dist (Distribution): Base distribution.
            transformer (Bijection): Bijection to be parameterised by conditioner.
            cond_dim (int, optional): Dimension of conditioning variables. Defaults to None.
            flow_layers (int, optional): Number of coupling layers. Defaults to 5.
            nn_width (int, optional): Conditioner hidden layer size. Defaults to 40.
            nn_depth (int, optional): Conditioner depth. Defaults to 2.
            nn_activation (int, optional): Conditioner activation function. Defaults to jnn.relu.
            invert: (bool, optional): Whether to invert the bijection. Broadly, True will prioritise a faster `inverse` methods, leading to faster `log_prob`, False will prioritise faster `transform` methods, leading to faster `sample`. Defaults to True
        """
        if base_dist.ndim != 1:
            raise ValueError(f"Expected base_dist.ndim==1, got {base_dist.ndim}")

        dim = base_dist.shape[0]
        permute_strategy = _default_permute_strategy(dim)

        def make_layer(key):  # coupling layer + permutation
            c_key, p_key = random.split(key)
            coupling = Coupling(
                key=c_key,
                transformer=transformer,
                d=dim // 2,
                D=dim,
                cond_dim=cond_dim,
                nn_width=nn_width,
                nn_depth=nn_depth,
                nn_activation=nn_activation,
            )

            if permute_strategy == "flip":
                return Chain([coupling, Flip()])
            elif permute_strategy == "random":
                perm = Permute(random.permutation(p_key, jnp.arange(dim)))
                return Chain([coupling, perm])
            else:
                return coupling

        keys = random.split(key, flow_layers)
        layers = eqx.filter_vmap(make_layer)(keys)
        bijection = Scan(layers)
        bijection = Invert(bijection) if invert else bijection

        self.nn_width = nn_width
        self.nn_depth = nn_depth
        self.flow_layers = flow_layers
        self.permute_strategy = permute_strategy
        self.shape = (dim,)
        self.cond_shape = None if cond_dim is None else (cond_dim,)
        super().__init__(base_dist, bijection)


class MaskedAutoregressiveFlow(Transformed):
    flow_layers: int
    nn_width: int
    nn_depth: int
    permute_strategy: str

    def __init__(
        self,
        key: KeyArray,
        base_dist: Distribution,
        transformer: Bijection,
        cond_dim: int = None,
        flow_layers: int = 8,
        nn_width: int = 40,
        nn_depth: int = 2,
        nn_activation: Callable = jnn.relu,
        invert: bool = True,
    ):
        """Masked autoregressive flow (https://arxiv.org/abs/1606.04934,
        https://arxiv.org/abs/1705.07057v4). Parameterises a transformer with an
        autoregressive neural network.

        Args:
            key (KeyArray): Random seed.
            base_dist (Distribution): Base distribution.
            transformer (Bijection): Bijection parameterised by autoregressive network.
            cond_dim (int, optional): _description_. Defaults to 0.
            flow_layers (int, optional): Number of flow layers. Defaults to 5.
            nn_width (int, optional): Number of hidden layers in neural network. Defaults to 40.
            nn_depth (int, optional): Depth of neural network. Defaults to 2.
            nn_activation (Callable, optional): _description_. Defaults to jnn.relu.
            invert (bool, optional): Whether to invert the bijection. Broadly, True will
                prioritise a faster inverse, leading to faster `log_prob`, False will prioritise
                faster forward, leading to faster `sample`. Defaults to True. Defaults to True.
        """
        if len(base_dist.shape) != 1:
            raise ValueError(f"Expected base_dist.ndim==1, got {base_dist.ndim}")

        dim = base_dist.shape[0]
        permute_strategy = _default_permute_strategy(dim)

        def make_layer(key):  # masked autoregressive layer + permutation
            masked_auto_key, p_key = random.split(key)
            masked_autoregressive = MaskedAutoregressive(
                key=masked_auto_key,
                transformer=transformer,
                dim=dim,
                cond_dim=cond_dim,
                nn_width=nn_width,
                nn_depth=nn_depth,
                nn_activation=nn_activation,
            )
            if permute_strategy == "flip":
                return Chain([masked_autoregressive, Flip()])
            elif permute_strategy == "random":
                perm = Permute(random.permutation(p_key, jnp.arange(dim)))
                return Chain([masked_autoregressive, perm])
            else:
                return masked_autoregressive

        keys = random.split(key, flow_layers)
        layers = eqx.filter_vmap(make_layer)(keys)
        bijection = Scan(layers)
        bijection = Invert(bijection) if invert else bijection

        self.nn_width = nn_width
        self.nn_depth = nn_depth
        self.flow_layers = flow_layers
        self.permute_strategy = permute_strategy
        self.shape = (dim,)
        self.cond_shape = None if cond_dim is None else (cond_dim,)
        super().__init__(base_dist, bijection)


class BlockNeuralAutoregressiveFlow(Transformed):
    flow_layers: int
    nn_block_dim: int
    nn_depth: int
    permute_strategy: str

    def __init__(
        self,
        key: KeyArray,
        base_dist: Distribution,
        cond_dim: Optional[int] = None,
        nn_depth: int = 1,
        nn_block_dim: int = 8,
        flow_layers: int = 1,
        invert: bool = True,
    ):
        """Block neural autoregressive flow (BNAF) (https://arxiv.org/abs/1904.04676).

        Args:
            key (KeyArray): Jax PRNGKey.
            base_dist (Distribution): Base distribution.
            cond_dim (Union[None, Tuple[int]]): Dimension of conditional variables.
            nn_depth (int, optional): Number of hidden layers within the networks. Defaults to 1.
            nn_block_dim (int, optional): Block size. Hidden layer width is dim*nn_block_dim. Defaults to 8.
            flow_layers (int, optional): Number of BNAF layers. Defaults to 1.
            invert: (bool, optional): Use `True` for access of `log_prob` only (e.g. fitting by maximum likelihood), `False` for the forward direction (sampling) only (e.g. for fitting variationally).
        """
        if len(base_dist.shape) != 1:
            raise ValueError(f"Expected base_dist.ndim==1, got {base_dist.ndim}")

        dim = base_dist.shape[-1]
        permute_strategy = _default_permute_strategy(base_dist.shape[0])

        def make_layer(key):  # masked autoregressive layer + permutation
            ban_key, p_key = random.split(key)
            ban = BlockAutoregressiveNetwork(
                key=ban_key,
                dim=dim,
                cond_dim=cond_dim,
                depth=nn_depth,
                block_dim=nn_block_dim,
            )
            if permute_strategy == "flip":
                return Chain([ban, Flip()])
            elif permute_strategy == "random":
                perm = Permute(random.permutation(p_key, jnp.arange(dim)))
                return Chain([ban, perm])
            else:
                return ban

        keys = random.split(key, flow_layers)
        layers = eqx.filter_vmap(make_layer)(keys)
        bijection = Scan(layers)
        bijection = Invert(bijection) if invert else bijection

        self.nn_block_dim = nn_block_dim
        self.nn_depth = nn_depth
        self.flow_layers = flow_layers
        self.permute_strategy = permute_strategy

        self.shape = (dim,)
        self.cond_shape = None if cond_dim is None else (cond_dim,)
        super().__init__(base_dist, bijection)


class TriangularSplineFlow(Transformed):
    flow_layers: int
    permute_strategy: str
    knots: int
    tanh_max_val: float

    def __init__(
        self,
        key: KeyArray,
        base_dist: Distribution,
        cond_dim: Optional[int] = None,
        flow_layers: int = 8,
        knots: int = 8,
        tanh_max_val: float = 3.0,
        invert: bool = True,
        permute_strategy: Optional[str] = None,
        init: Callable = glorot_uniform(),
    ):

        if len(base_dist.shape) != 1:
            raise ValueError(f"Expected base_dist.ndim==1, got {base_dist.ndim}")

        dim = base_dist.shape[-1]

        permute_strategy = _default_permute_strategy(dim)

        def make_layer(key):

            lt_key, perm_key = random.split(key)
            c_dim = 0 if cond_dim is None else cond_dim
            weights = init(lt_key, (dim, dim + c_dim))
            lt_weights = weights[:, :dim].at[jnp.diag_indices(dim)].set(1)
            cond_weights = weights[:, dim:]
            lt = TriangularAffine(jnp.zeros(dim), lt_weights, weight_normalisation=True)

            bijections = [
                TanhLinearTails(tanh_max_val),
                RationalQuadraticSpline(knots, interval=1, shape=(dim,)),
                Invert(TanhLinearTails(tanh_max_val)),
                lt,
            ]

            if cond_dim is not None:
                linear_condition = AdditiveLinearCondition(cond_weights)
                bijections.append(linear_condition)

            if permute_strategy == "flip":
                bijections.append(Flip())
            elif permute_strategy == "random":
                bijections.append(
                    Permute(random.permutation(perm_key, jnp.arange(dim)))
                )
            return Chain(bijections)

        keys = random.split(key, flow_layers)
        layers = eqx.filter_vmap(make_layer)(keys)
        bijection = Scan(layers)
        bijection = Invert(bijection) if invert else bijection

        self.flow_layers = flow_layers
        self.permute_strategy = permute_strategy
        self.knots = knots
        self.tanh_max_val = tanh_max_val
        self.shape = (dim,)
        self.cond_shape = None if cond_dim is None else (cond_dim,)

        super().__init__(base_dist, bijection)


def _default_permute_strategy(dim):
    if dim <= 2:
        return {1: "none", 2: "flip"}[dim]
    else:
        return "random"



from abc import ABC, abstractmethod
from typing import Optional, Tuple, Union

import equinox as eqx
import jax.numpy as jnp

import jax.random as jr
from jax.scipy import stats as jstats

from flowjax.bijections import Affine, Bijection
from flowjax.utils import Array, _get_ufunc_signature
from math import prod
from flowjax.utils import merge_shapes

from jax.experimental import checkify

class Distribution(eqx.Module, ABC):
    """Distribution base class. Distributions all have an attribute ``shape``,
    denoting the shape of a single sample from the distribution. This corresponds to the 
    ``batch_shape + event_shape`` in torch/numpyro distributions. Similarly, the
    ``cond_shape`` attribute denotes the shape of the conditioning variable.
    This attribute is None for unconditional distributions. For example

    .. doctest::

        >>> import jax.numpy as jnp
        >>> from flowjax.distributions import Normal
        >>> dist = Normal(jnp.zeros(2))
        >>> dist.shape
        (2,)
        >>> dist.cond_shape is None
        True

    Distributions are registered as jax PyTrees (as they are equinox modules), and as such
    they are compatible with normal jax operations.

    Implementing a distribution

        (1) Inherit from ``Distribution``.
        (2) Define attributes ``shape`` and ``cond_shape`` (cond shape should be None for unconditional distributions).
        (3) Define the ``_sample`` method, which samples a point with a shape of ``shape``, (given a conditioning variable with shape ``cond_shape`` for conditional distributions).
        (4) Define the ``_log_prob`` method, which evaluates the log probability, given an input of shape ``shape`` (and a conditioning variable with shape ``cond_shape`` for conditional distributions).

        The base class will handle defining more convenient log_prob and sample methods that support broadcasting and perform argument checks.

    """

    shape: Tuple[int]
    cond_shape: Union[None, Tuple[int]]

    @abstractmethod
    def _log_prob(self, x: Array, condition: Optional[Array] = None):
        "Evaluate the log probability of point x."
        pass

    @abstractmethod
    def _sample(self, key: jr.KeyArray, condition: Optional[Array] = None):
        "Sample a point from the distribution."
        pass

    def _sample_and_log_prob(
        self,
        key: jr.PRNGKey,
        condition: Optional[Array] = None
        ):
        """
        Sample a point from the distribution, and return its log probability.
        Subclasses can reimplement this method in cases where more efficient methods exists (e.g. see Transformed).
        """
        x = self._sample(key, condition)
        log_prob = self._log_prob(x, condition)
        return x, log_prob

    def log_prob(self, x: Array, condition: Optional[Array] = None):
        """Evaluate the log probability. Uses numpy like broadcasting if additional
        leading dimensions are passed.

        Args:
            x (Array): Points at which to evaluate density.
            condition (Optional[Array], optional): Conditioning variables. Defaults to None.

        Returns:
            Array: Jax array of log probabilities.
        """
        self._argcheck(x, condition)
        if condition is not None:
            sig = _get_ufunc_signature([self.shape, self.cond_shape], [()])
            exclude = {}
        else:
            sig = _get_ufunc_signature([self.shape], [()])
            exclude = {1}
        lps = jnp.vectorize(self._log_prob, signature=sig, excluded=exclude)(
            x, condition
        )
        return jnp.where(jnp.isnan(lps), -jnp.inf, lps)

    def sample(
        self,
        key: jr.PRNGKey,
        sample_shape: Tuple[int] = (),
        condition: Optional[Array] = None,
    ):
        """Sample from the distribution. For unconditional distributions, the output will
        be of shape ``sample_shape + dist.shape``.
        
        Example:

            .. testsetup::

                from flowjax.distributions import StandardNormal
                import jax.random as jr
                import jax.numpy as jnp
                from flowjax.flows import CouplingFlow
                from flowjax.bijections import Affine
                # For a unconditional distribution:
                key = jr.PRNGKey(0)
                dist = StandardNormal((2,))
                # For a conditional distribution
                cond_dist = CouplingFlow(key, StandardNormal((2,)), cond_dim=3, transformer=Affine())

            For an unconditional distribution:

            .. doctest::

                >>> dist.shape
                (2,)
                >>> samples = dist.sample(key, (10, ))
                >>> samples.shape
                (10, 2)

            For a conditional distribution:

            .. doctest::

                >>> cond_dist.shape
                (2,)
                >>> cond_dist.cond_shape
                (3,)
                >>> # Sample 10 times for a particular condition
                >>> samples = cond_dist.sample(key, (10,), condition=jnp.ones(3))
                >>> samples.shape
                (10, 2)
                >>> # Sampling, batching over a condition
                >>> samples = cond_dist.sample(key, condition=jnp.ones((5, 3)))
                >>> samples.shape
                (5, 2)
                >>> # Sample 10 times for each of 5 conditioning variables
                >>> samples = cond_dist.sample(key, (10,), condition=jnp.ones((5, 3)))
                >>> samples.shape
                (10, 5, 2)

        Args:
            key (jr.PRNGKey): Jax random key.
            condition (Optional[Array], optional): Conditioning variables. Defaults to None.
            sample_shape (Tuple[int], optional): Sample shape. Defaults to ().

        """
        self._argcheck(condition=condition)

        if condition is None:
            key_shape = sample_shape
            excluded = {1}
            sig = _get_ufunc_signature([(2,)], [self.shape])
        else:
            leading_cond_shape = condition.shape[: -len(self.cond_shape)] if self.cond_ndim > 0 else condition.shape
            key_shape = sample_shape + leading_cond_shape
            excluded = {}
            sig = _get_ufunc_signature([(2,), self.cond_shape], [self.shape])

        key_size = max(1, prod(key_shape))  # Still need 1 key for scalar input
        keys = jnp.reshape(jr.split(key, key_size), key_shape + (2,))

        return jnp.vectorize(self._sample, excluded=excluded, signature=sig)(
            keys, condition
        )


    def sample_and_log_prob(
        self,
        key: jr.PRNGKey,
        sample_shape: Tuple[int] = (),
        condition: Optional[Array] = None,
        ):
        """Sample the distribution and return the samples and corresponding log probabilities.
        For transformed distributions (especially flows), this will generally be more efficient
        than calling the methods seperately.
        
        Refer to the :py:meth:`~flowjax.distributions.Distribution.sample` and
        Refer to the :py:meth:`~flowjax.distributions.Distribution.log_prob` documentation
        for more information.

        Args:
            key (jr.PRNGKey): Jax random key.
            condition (Optional[Array], optional): Conditioning variables. Defaults to None.
            sample_shape (Tuple[int], optional): Sample shape. Defaults to ().
        """
        self._argcheck(condition=condition)

        if condition is None:
            key_shape = sample_shape
            excluded = {1}
            sig = _get_ufunc_signature([(2,)], [self.shape, ()])
        else:
            leading_cond_shape = condition.shape[: -len(self.cond_shape)] if self.cond_ndim > 0 else condition.shape
            key_shape = sample_shape + leading_cond_shape
            excluded = {}
            sig = _get_ufunc_signature([(2,), self.cond_shape], [self.shape, ()])

        key_size = max(1, prod(key_shape))  # Still need 1 key for scalar input
        keys = jnp.reshape(jr.split(key, key_size), key_shape + (2,))

        return jnp.vectorize(self._sample_and_log_prob, excluded=excluded, signature=sig)(
            keys, condition
        )

    def _argcheck(self, x=None, condition=None):
        # jnp.vectorize would catch ndim mismatches, but it doesn't check axis lengths.
        if x is not None:
            x_trailing = x.shape[-self.ndim :] if self.ndim > 0 else ()
            if x_trailing != self.shape:
                raise ValueError(
                    f"Expected trailing dimensions in input x to match the distribution shape, but got "
                    f"x shape {x.shape}, and distribution shape {self.shape}."
                )

        if condition is None and self.cond_shape is not None:
            raise ValueError(
                f"Conditioning variable was not provided. "
                f"Expected conditioning variable with trailing shape {self.shape}."
            )

        if condition is not None:
            if self.cond_shape is None:
                raise ValueError("condition should not be provided for unconditional distribution.")
            else:
                condition_trailing = (
                    condition.shape[-self.cond_ndim :] if self.cond_ndim > 0 else ()
                )
                if condition_trailing != self.cond_shape:
                    raise ValueError(
                        f"Expected trailing dimensions in the condition to match distribution.cond_shape, but got "
                        f"condition shape {condition.shape}, and distribution.cond_shape {self.cond_shape}."
                    )

    @property
    def ndim(self):
        return len(self.shape)

    @property
    def cond_ndim(self):
        return len(self.cond_shape)


class Transformed(Distribution):
    base_dist: Distribution
    bijection: Bijection
    cond_shape: Union[None, Tuple[int]]

    def __init__(
        self,
        base_dist: Distribution,
        bijection: Bijection,
    ):
        """
        Form a distribution like object using a base distribution and a
        bijection. We take the forward bijection for use in sampling, and the inverse
        bijection for use in density evaluation.

        Args:
            base_dist (Distribution): Base distribution.
            bijection (Bijection): Bijection to transform distribution.

        Example:

        .. doctest::

            >>> from flowjax.distributions import StandardNormal, Transformed
            >>> from flowjax.bijections import Affine
            >>> normal = StandardNormal()
            >>> bijection = Affine(1)
            >>> transformed = Transformed(normal, bijection)

        
        .. warning::
            It is the currently the users responsibility to ensure the bijection is valid
            across the entire support of the distribution. Failure to do so may lead to
            to unexpected results. In future versions explicit constraints may be introduced.
        """
        self.base_dist = base_dist
        self.bijection = bijection
        self.shape = self.base_dist.shape
        self.cond_shape = merge_shapes(
            (self.bijection.cond_shape, self.base_dist.cond_shape)
        )

    def _log_prob(self, x: Array, condition: Optional[Array] = None):
        z, log_abs_det = self.bijection.inverse_and_log_abs_det_jacobian(x, condition)
        p_z = self.base_dist._log_prob(z, condition)
        return p_z + log_abs_det

    def _sample(self, key: jr.KeyArray, condition: Optional[Array] = None):
        z = self.base_dist._sample(key, condition)
        x = self.bijection.transform(z, condition)
        return x

    def _sample_and_log_prob(self, key: jr.PRNGKey, condition: Optional[Array] = None):
        # We overwrite the naive implementation of calling both methods seperately to
        # avoid computing the inverse transformation.
        x, log_prob_base = self.base_dist._sample_and_log_prob(key, condition)
        y, forward_log_dets = self.bijection.transform_and_log_abs_det_jacobian(x, condition)
        return y, log_prob_base - forward_log_dets


class StandardNormal(Distribution):
    def __init__(self, shape: Tuple[int] = ()):
        """
        Implements a standard normal distribution, condition is ignored.

        Args:
            shape (Tuple[int]): The shape of the normal distribution. Defaults to ().
        """
        self.shape = shape
        self.cond_shape = None

    def _log_prob(self, x: Array, condition: Optional[Array] = None):
        return jstats.norm.logpdf(x).sum()

    def _sample(self, key: jr.KeyArray, condition: Optional[Array] = None):
        return jr.normal(key, self.shape)


class Normal(Transformed):
    """
    Implements an independent Normal distribution with mean and std for
    each dimension. `loc` and `scale` should be broadcastable.
    """

    def __init__(self, loc: Array=0, scale: Array=1):
        """
        Args:
            loc (Array): Means. Defaults to 0.
            scale (Array): Standard deviations. Defaults to 1.
        """
        self.shape = jnp.broadcast_shapes(jnp.shape(loc), jnp.shape(scale))
        self.cond_shape = None
        base_dist = StandardNormal(self.shape)
        bijection = Affine(loc=loc, scale=scale)
        super().__init__(base_dist, bijection)

    @property
    def loc(self):
        return self.bijection.loc

    @property
    def scale(self):
        return self.bijection.scale


class _StandardUniform(Distribution):
    """
    Implements a standard independent Uniform distribution, ie X ~ Uniform([0, 1]^dim).
    """

    def __init__(self, shape: Tuple[int] = ()):
        self.shape = shape
        self.cond_shape = None

    def _log_prob(self, x: Array, condition: Optional[Array] = None):
        return jstats.uniform.logpdf(x).sum()

    def _sample(self, key: jr.KeyArray, condition: Optional[Array] = None):
        return jr.uniform(key, shape=self.shape)




class Uniform(Transformed):
    """
    Implements an independent uniform distribution
    between min and max for each dimension. `minval` and `maxval` should be broadcastable.
    """

    def __init__(self, minval: Array, maxval: Array):
        """
        Args:
            minval (Array): Minimum values.
            maxval (Array): Maximum values.
        """
        self.shape = jnp.broadcast_shapes(jnp.shape(minval), jnp.shape(maxval))
        self.cond_shape = None

        checkify.check(
            jnp.all(maxval >= minval), "Minimums must be less than the maximums."
        )

        base_dist = _StandardUniform(self.shape)
        bijection = Affine(loc=minval, scale=maxval - minval)
        super().__init__(base_dist, bijection)

    @property
    def minval(self):
        return self.bijection.loc

    @property
    def maxval(self):
        return self.bijection.loc + self.bijection.scale


class _StandardGumbel(Distribution):
    """Standard gumbel distribution (https://en.wikipedia.org/wiki/Gumbel_distribution)."""

    def __init__(self, shape: Tuple[int] = ()):
        self.shape = shape
        self.cond_shape = None

    def _log_prob(self, x: Array, condition: Optional[Array] = None):
        return -(x + jnp.exp(-x)).sum()

    def _sample(self, key: jr.KeyArray, condition: Optional[Array] = None):
        return jr.gumbel(key, shape=self.shape)


class Gumbel(Transformed):
    """Gumbel distribution (https://en.wikipedia.org/wiki/Gumbel_distribution)"""

    def __init__(self, loc: Array=0, scale: Array=1):
        """
        `loc` and `scale` should broadcast to the dimension of the distribution.

        Args:
            loc (Array): Location paramter.
            scale (Array, optional): Scale parameter. Defaults to 1.0.
        """
        self.shape = jnp.broadcast_shapes(jnp.shape(loc), jnp.shape(scale))
        self.cond_shape = None
        base_dist = _StandardGumbel(self.shape)
        bijection = Affine(loc, scale)

        super().__init__(base_dist, bijection)

    @property
    def loc(self):
        return self.bijection.loc

    @property
    def scale(self):
        return self.bijection.scale


class _StandardCauchy(Distribution):
    """
    Implements standard cauchy distribution (loc=0, scale=1)
    Ref: https://en.wikipedia.org/wiki/Cauchy_distribution
    """

    def __init__(self, shape: Tuple[int] = ()):
        self.shape = shape
        self.cond_shape = None

    def _log_prob(self, x: Array, condition: Optional[Array] = None):
        return jstats.cauchy.logpdf(x).sum()

    def _sample(self, key: jr.KeyArray, condition: Optional[Array] = None):
        return jr.cauchy(key, shape=self.shape)


class Cauchy(Transformed):
    """
    Cauchy distribution (https://en.wikipedia.org/wiki/Cauchy_distribution).
    """

    def __init__(self, loc: Array=0, scale: Array=1):
        """
        `loc` and `scale` should broadcast to the dimension of the distribution.

        Args:
            loc (Array): Location paramter.
            scale (Array, optional): Scale parameter. Defaults to 1.0.
        """
        self.shape = jnp.broadcast_shapes(jnp.shape(loc), jnp.shape(scale))
        self.cond_shape = None
        base_dist = _StandardCauchy(self.shape)
        bijection = Affine(loc, scale)
        super().__init__(base_dist, bijection)

    @property
    def loc(self):
        return self.bijection.loc

    @property
    def scale(self):
        return self.bijection.scale


class _StandardStudentT(Distribution):
    """
    Implements student T distribution with specified degrees of freedom.
    """

    log_df: Array

    def __init__(self, df: Array):
        self.shape = df.shape
        self.cond_shape = None
        self.log_df = jnp.log(df)

    def _log_prob(self, x: Array, condition: Optional[Array] = None):
        return jstats.t.logpdf(x, df=self.df).sum()

    def _sample(self, key: jr.KeyArray, condition: Optional[Array] = None):
        return jr.t(key, df=self.df, shape=self.shape)

    @property
    def df(self):
        return jnp.exp(self.log_df)


class StudentT(Transformed):
    """Student T distribution (https://en.wikipedia.org/wiki/Student%27s_t-distribution)."""

    def __init__(self, df: Array, loc: Array=0, scale: Array=1):
        """
        `df`, `loc` and `scale` broadcast to the dimension of the distribution.

        Args:
            df (Array): The degrees of freedom.
            loc (Array): Location parameter. Defaults to 0.0.
            scale (Array, optional): Scale parameter. Defaults to 1.0.
        """
        df, loc, scale = jnp.broadcast_arrays(df, loc, scale)
        self.shape = df.shape
        self.cond_shape = None
        base_dist = _StandardStudentT(df)
        bijection = Affine(loc, scale)
        super().__init__(base_dist, bijection)

    @property
    def loc(self):
        return self.bijection.loc

    @property
    def scale(self):
        return self.bijection.scale

    @property
    def df(self):
        return self.base_dist.df



from typing import Sequence, List
from flowjax.utils import Array
import jax.random as jr
import jax.numpy as jnp


def train_val_split(key: jr.KeyArray, arrays: Sequence[Array], val_prop: float = 0.1):
    """Train validation split along axis 0.

    Args:
        key (KeyArray): Jax PRNGKey
        arrays List[Array]: List of arrays.
        val_prop (float, optional): Proportion of data to use for validation. Defaults to 0.1.

    Returns:
        Tuple[Tuple]: (train_arrays, validation_arrays)
    """
    if not (0 <= val_prop <= 1):
        raise ValueError("val_prop should be between 0 and 1.")
    n = arrays[0].shape[0]
    permutation = jr.permutation(key, jnp.arange(n))
    arrays = tuple(a[permutation] for a in arrays)
    n_train = n - round(val_prop * n)
    train = tuple(a[:n_train] for a in arrays)
    val = tuple(a[n_train:] for a in arrays)
    return train, val


def count_fruitless(losses: List[float]) -> int:
    """Given a list of losses from each epoch, count the number of epochs since
    the minimum loss.

    Args:
        losses (List[float]): List of losses.

    """
    min_idx = jnp.argmin(jnp.array(losses)).item()
    return len(losses) - min_idx - 1



from typing import Sequence, Tuple, Union

from flowjax.utils import merge_shapes
from flowjax.bijections import Bijection
from flowjax.utils import Array


class Chain(Bijection):
    """Chain together arbitrary bijections to form another bijection."""

    bijections: Tuple[Bijection]

    def __init__(self, bijections: Sequence[Bijection]):
        """
        Args:
            bijections (Sequence[Bijection]): Sequence of bijections.
        """
        self.shape = merge_shapes([s.shape for s in bijections])
        self.cond_shape = merge_shapes([b.cond_shape for b in bijections])
        self.bijections = tuple(bijections)

    def transform(self, x, condition=None):
        for bijection in self.bijections:
            x = bijection.transform(x, condition)
        return x

    def transform_and_log_abs_det_jacobian(self, x, condition=None):
        log_abs_det_jac = 0
        for bijection in self.bijections:
            x, log_abs_det_jac_i = bijection.transform_and_log_abs_det_jacobian(
                x, condition
            )
            log_abs_det_jac += log_abs_det_jac_i.sum()
        return x, log_abs_det_jac

    def inverse(self, y: Array, condition=None):
        for bijection in reversed(self.bijections):
            y = bijection.inverse(y, condition)
        return y

    def inverse_and_log_abs_det_jacobian(self, y, condition=None):
        log_abs_det_jac = 0
        for bijection in reversed(self.bijections):
            y, log_abs_det_jac_i = bijection.inverse_and_log_abs_det_jacobian(
                y, condition
            )
            log_abs_det_jac += log_abs_det_jac_i.sum()
        return y, log_abs_det_jac

    def __getitem__(self, i: Union[int, slice]) -> Bijection:
        if isinstance(i, int):
            return self.bijections[i]
        elif isinstance(i, slice):
            return Chain(self.bijections[i])
        else:
            raise TypeError(f"Indexing with type {type(i)} is not supported.")

    def __iter__(self):
        yield from self.bijections

    def __len__(self):
        return len(self.bijections)



from typing import Optional

import jax.numpy as jnp
from jax.experimental import checkify
from jax.scipy.linalg import solve_triangular

from flowjax.bijections import Bijection
from flowjax.utils import Array
from jax.experimental import checkify


class Affine(Bijection):
    loc: Array
    log_scale: Array
    
    def __init__(self, loc: Array=0, scale: Array=1):
        """Elementwise affine transformation y = ax + b. loc and scale should broadcast
        to the desired shape of the bijection.

        Args:
            loc (int, optional): Location parameter. Defaults to 0.
            scale (int, optional): Scale parameter. Defaults to 1.
        """
        loc, scale = [jnp.asarray(a, dtype=jnp.float32) for a in (loc, scale)]
        self.shape = jnp.broadcast_shapes(jnp.shape(loc), jnp.shape(scale))
        self.cond_shape = None

        self.loc = jnp.broadcast_to(loc, self.shape)
        self.log_scale = jnp.broadcast_to(jnp.log(scale), self.shape)

    def transform(self, x, condition=None):
        self._argcheck(x)
        return x * self.scale + self.loc

    def transform_and_log_abs_det_jacobian(self, x, condition=None):
        self._argcheck(x)
        return x * self.scale + self.loc, self.log_scale.sum()

    def inverse(self, y, condition=None):
        self._argcheck(y)
        return (y - self.loc) / self.scale

    def inverse_and_log_abs_det_jacobian(self, y, condition=None):
        self._argcheck(y)
        return (y - self.loc) / self.scale, -self.log_scale.sum()

    @property
    def scale(self):
        return jnp.exp(self.log_scale)


class TriangularAffine(Bijection):
    """Transformation of the form ``Ax + b``, where ``A`` is a lower or upper triangular matrix."""

    loc: Array
    diag_idxs: Array
    tri_mask: Array
    lower: bool
    weight_log_scale: Optional[Array]
    _arr: Array
    _log_diag: Array

    def __init__(
        self,
        loc: Array,
        arr: Array,
        lower: bool = True,
        weight_normalisation: bool = False,
    ):
        """
        Args:
            loc (Array): Location parameter.
            arr (Array): Triangular matrix.
            lower (bool, optional): Whether the mask should select the lower or upper triangular matrix (other elements ignored). Defaults to True.
            weight_log_scale (Optional[Array], optional): If provided, carry out weight normalisation.
        """

        if (arr.ndim != 2) or (arr.shape[0] != arr.shape[1]):
            ValueError("arr must be a square, 2-dimensional matrix.")
        checkify.check(
            jnp.all(jnp.diag(arr) > 0),
            "arr diagonal entries must be greater than 0",
        )
        dim = arr.shape[0]
        self.diag_idxs = jnp.diag_indices(dim)
        tri_mask = jnp.tril(jnp.ones((dim, dim), dtype=jnp.int32), k=-1)
        self.tri_mask = tri_mask if lower else tri_mask.T
        self.lower = lower

        # inexact arrays
        self.loc = jnp.broadcast_to(loc, (dim,))
        self._arr = arr
        self._log_diag = jnp.log(jnp.diag(arr))
        self.weight_log_scale = jnp.zeros((dim, 1)) if weight_normalisation else None

        self.shape = (dim,)
        self.cond_shape = None

    @property
    def arr(self):
        "Get triangular array, (applies masking and min_diag constraint)."
        diag = jnp.exp(self._log_diag)
        off_diag = self.tri_mask * self._arr
        arr = off_diag.at[self.diag_idxs].set(diag)

        if self.weight_log_scale is not None:
            norms = jnp.linalg.norm(arr, axis=1, keepdims=True)
            arr = jnp.exp(self.weight_log_scale) * arr / norms

        return arr

    def transform(self, x, condition=None):
        self._argcheck(x)
        return self.arr @ x + self.loc

    def transform_and_log_abs_det_jacobian(self, x, condition=None):
        self._argcheck(x)
        a = self.arr
        return a @ x + self.loc, jnp.log(jnp.diag(a)).sum()

    def inverse(self, y, condition=None):
        self._argcheck(y)
        return solve_triangular(self.arr, y - self.loc, lower=self.lower)

    def inverse_and_log_abs_det_jacobian(self, y, condition=None):
        self._argcheck(y)
        a = self.arr
        x = solve_triangular(a, y - self.loc, lower=self.lower)
        return x, -jnp.log(jnp.diag(a)).sum()


class AdditiveLinearCondition(Bijection):
    """Carries out ``y = x + W @ condition``, as the forward transformation and
    ``x = y - W @ condition`` as the inverse."""

    W: Array

    def __init__(self, arr: Array):
        """
        Args:
            arr (Array): Array (``W`` in the description.)
        """
        self.W = arr
        super().__init__(shape=(arr.shape[-2],), cond_shape=(arr.shape[-1],))

    def transform(self, x, condition=None):
        self._argcheck(x, condition)
        return x + self.W @ condition

    def transform_and_log_abs_det_jacobian(self, x, condition=None):
        self._argcheck(x, condition)
        return self.transform(x, condition), jnp.array(0)

    def inverse(self, y, condition=None):
        self._argcheck(y, condition)
        return y - self.W @ condition

    def inverse_and_log_abs_det_jacobian(self, y, condition=None):
        self._argcheck(y, condition)
        return self.inverse(y, condition), jnp.array(0)




from flowjax.bijections import Bijection
import jax.numpy as jnp

class Exp(Bijection):
    def __init__(self):
        """Elementwise exponential transform (forward) and log transform (inverse)."""
        self.shape = None
        self.cond_shape = None

    def transform(self, x, condition = None):
        return jnp.exp(x)

    def transform_and_log_abs_det_jacobian(self, x, condition = None):
        return jnp.exp(x), x.sum()

    def inverse(self, y, condition = None):
        return jnp.log(y)

    def inverse_and_log_abs_det_jacobian(self, y, condition = None):
        x = jnp.log(y)
        return x, -x.sum()



from typing import Callable

import jax.random as jr
import optax
import equinox as eqx
from tqdm import tqdm

from flowjax.utils import Array
from flowjax.distributions import Distribution
from jaxtyping import PyTree
from equinox.custom_types import BoolAxisSpec

@eqx.filter_jit
def elbo_loss(key: jr.KeyArray, dist: Distribution, target: Callable[[Array], Array], num_samples: int):
    samples, approx_density = dist.sample_and_log_prob(key, (num_samples,))
    target_density = target(samples)
    losses = approx_density - target_density
    return losses.mean()

def fit_to_variational_target(
    key: jr.KeyArray,
    dist: Distribution,
    target: Callable[[Array], Array],
    loss_fn: Callable[[Distribution, Callable, jr.KeyArray, int], float] = elbo_loss,
    steps: int = 100,
    samples_per_step: int = 500,
    learning_rate: float = 5e-4,
    clip_norm: float = 0.5,
    optimizer: optax.GradientTransformation = None,
    filter_spec: PyTree[BoolAxisSpec] = eqx.is_inexact_array,
    show_progress: bool = True,
):
    """
    Train a distribution (e.g. a flow) by variational inference to a target
    (e.g. an unnormalized density).
    
    Args:
        key (KeyArray): Jax PRNGKey.
        dist (Distribution): Distribution object, trainable parameters are found using equinox.is_inexact_array.
        target (Callable): The variational target (this is usually the unormalized log posterior)
        loss_fn (Callable, optional): The loss function to optimize. The loss function should take
            a random key, the distribution, a callable that maps samples from the distribution to a scalar loss,
            and a number of samples to use. Defaults to elbo_loss.
        steps (int, optional): The number of training steps to run. Defaults to 100.
        samples_per_step (int, optional): number of samples to use at each step. Defaults to 500.
        learning_rate (float, optional): Adam learning rate. Defaults to 5e-4.
        clip_norm (float, optional): Maximum gradient norm before clipping occurs. Defaults to 0.5.
        optimizer (optax.GradientTransformation): Optax optimizer. If provided, this overrides the default Adam optimizer, and the learning_rate and clip_norm arguments are ignored. Defaults to None.
        filter_spec (PyTree[BoolAxisSpec], optional): Equinox `filter_spec` for specifying trainable parameters. Either a callable `leaf -> bool`, or a PyTree with prefix structure matching `dist` with True/False values. Defaults to `eqx.is_inexact_array`.
        show_progress (bool, optional): Whether to show progress bar. Defaults to True.
    """
    if optimizer is None:
        optimizer = optimizer = optax.chain(
            optax.clip_by_global_norm(clip_norm), optax.adam(learning_rate=learning_rate)
        )

    @eqx.filter_jit
    def step(dist, target, key, optimizer, opt_state):
        loss_val, grads = eqx.filter_value_and_grad(loss_fn, arg=filter_spec)(
            key, dist, target, samples_per_step
            )
        updates, opt_state = optimizer.update(grads, opt_state)
        dist = eqx.apply_updates(dist, updates)
        return dist, opt_state, loss_val

    trainable_params, _ = eqx.partition(dist, filter_spec)
    opt_state = optimizer.init(trainable_params)

    losses = []

    loop = tqdm(range(steps)) if show_progress is True else range(steps)
    for iteration in loop:
        key, subkey = jr.split(key)
        dist, opt_state, loss = step(dist, target, subkey, optimizer, opt_state)
        
        losses.append(loss.item())

        if show_progress:
            loop.set_postfix({'loss': losses[-1]})

    return dist, losses


from typing import Dict, List, Optional

import equinox as eqx
import jax.numpy as jnp
import optax
from equinox.custom_types import BoolAxisSpec
from jax import random
from jax.random import KeyArray
from jaxtyping import PyTree
from tqdm import tqdm
import optax

from flowjax.distributions import Distribution
from flowjax.utils import Array

from flowjax.train.train_utils import train_val_split, count_fruitless

def fit_to_data(
    key: KeyArray,
    dist: Distribution,
    x: Array,
    condition: Optional[Array] = None,
    max_epochs: int = 50,
    max_patience: int = 5,
    batch_size: int = 256,
    val_prop: float = 0.1,
    learning_rate: float = 5e-4,
    clip_norm: float = 0.5,
    optimizer: Optional[optax.GradientTransformation] = None,
    filter_spec: PyTree[BoolAxisSpec] = eqx.is_inexact_array,
    show_progress: bool = True,
):
    """Train a distribution (e.g. a flow) to samples by maximum likelihood. Note that the last batch in each epoch is dropped if truncated.

    Args:
        key (KeyArray): Jax PRNGKey.
        dist (Distribution): Distribution object.
        x (Array): Samples from target distribution.
        condition (Optional[Array], optional): Conditioning variables. Defaults to None.
        max_epochs (int, optional): Maximum number of epochs. Defaults to 50.
        max_patience (int, optional): Number of consecutive epochs with no validation loss improvement after which training is terminated. Defaults to 5.
        batch_size (int, optional): Batch size. Defaults to 256.
        val_prop (float, optional): Proportion of data to use in validation set. Defaults to 0.1.
        learning_rate (float, optional): Adam learning rate. Defaults to 5e-4.
        clip_norm (float, optional): Maximum gradient norm before clipping occurs. Defaults to 0.5.
        optimizer (optax.GradientTransformation): Optax optimizer. If provided, this overrides the default Adam optimizer, and the learning_rate and clip_norm arguments are ignored. Defaults to None.
        filter_spec (PyTree[BoolAxisSpec], optional): Equinox `filter_spec` for specifying trainable parameters. Either a callable `leaf -> bool`, or a PyTree with prefix structure matching `dist` with True/False values. Defaults to `eqx.is_inexact_array`.
        show_progress (bool, optional): Whether to show progress bar. Defaults to True.
    """

    @eqx.filter_jit
    def loss_fn(dist, x, condition=None):
        return -dist.log_prob(x, condition).mean()

    @eqx.filter_jit
    def step(dist, optimizer, opt_state, x, condition=None):
        loss_val, grads = eqx.filter_value_and_grad(loss_fn, arg=filter_spec)(
            dist, x, condition
        )
        updates, opt_state = optimizer.update(grads, opt_state)
        dist = eqx.apply_updates(dist, updates)
        return dist, opt_state, loss_val

    if optimizer is None:
        optimizer = optimizer = optax.chain(
            optax.clip_by_global_norm(clip_norm), optax.adam(learning_rate=learning_rate)
        )
    
    best_params, static = eqx.partition(dist, filter_spec)
    opt_state = optimizer.init(best_params)

    key, train_val_split_key = random.split(key)

    inputs = (x,) if condition is None else (x, condition)
    train_args, val_args = train_val_split(train_val_split_key, inputs, val_prop=val_prop)
    train_len, val_len = train_args[0].shape[0], val_args[0].shape[0]
    if batch_size > train_len:
        raise ValueError(
            f"The batch size ({batch_size}) cannot be greater than the train set size ({train_len})."
            )

    losses = {"train": [], "val": []}  # type: Dict[str, List[float]]

    loop = tqdm(range(max_epochs)) if show_progress is True else range(max_epochs)

    for epoch in loop:
        key, subkey = random.split(key)
        permutation = random.permutation(subkey, jnp.arange(train_len))
        train_args = tuple(a[permutation] for a in train_args)

        epoch_train_loss = 0
        batch_start_idxs = range(0, train_len - batch_size + 1, batch_size)
        for i in batch_start_idxs:
            batch = tuple(a[i : i + batch_size] for a in train_args)
            dist, opt_state, loss_i = step(dist, optimizer, opt_state, *batch)
            epoch_train_loss += loss_i.item() / len(batch_start_idxs)

        epoch_val_loss = 0
        batch_start_idxs = range(0, val_len - batch_size + 1, batch_size)
        for i in batch_start_idxs:
            batch = tuple(a[i : i + batch_size] for a in val_args)
            epoch_val_loss += loss_fn(dist, *batch).item() / len(batch_start_idxs)

        losses["train"].append(epoch_train_loss)
        losses["val"].append(epoch_val_loss)

        if epoch_val_loss == min(losses["val"]):
            best_params = eqx.filter(dist, filter_spec)

        elif count_fruitless(losses["val"]) > max_patience:
            if show_progress == True:
                loop.set_postfix_str(f"{loop.postfix} (Max patience reached)")
            break

        if show_progress:
            loop.set_postfix({k: v[-1] for k, v in losses.items()})

    dist = eqx.combine(best_params, static)
    return dist, losses




from .block_autoregressive import BlockAutoregressiveLinear, BlockTanh
from .masked_autoregressive import AutoregressiveMLP, MaskedLinear

__all__ = [
    "MaskedLinear",
    "AutoregressiveMLP",
    "BlockAutoregressiveLinear",
    "BlockTanh",
]



"""Masked autoregressive network and bijection."""

from functools import partial
from typing import Callable, Union

import jax
import jax.nn as jnn
import jax.numpy as jnp
from jax.random import KeyArray
import equinox as eqx

from flowjax.bijections import Bijection
from flowjax.nn import AutoregressiveMLP
from flowjax.utils import Array, tile_until_length, get_ravelled_bijection_constructor

from flowjax.bijections.jax_transforms import Vmap


class MaskedAutoregressive(Bijection):
    transformer_constructor: Callable
    autoregressive_mlp: AutoregressiveMLP

    def __init__(
        self,
        key: KeyArray,
        transformer: Bijection,
        dim: int,
        cond_dim: Union[None, int],
        nn_width: int,
        nn_depth: int,
        nn_activation: Callable = jnn.relu,
    ) -> None:
        """Masked autoregressive bijection implementation (https://arxiv.org/abs/1705.07057v4).
        The transformer is parameterised by a neural network, with weights masked to ensure
        an autoregressive structure.

        Args:
            key (KeyArray): Jax PRNGKey
            transformer (Bijection): Bijection with shape () to be parameterised by the autoregressive network.
            dim (int): Dimension.
            cond_dim (Union[None, int]): Dimension of any conditioning variables.
            nn_width (int): Neural network width.
            nn_depth (int): Neural network depth.
            nn_activation (Callable, optional): Neural network activation. Defaults to jnn.relu.
        """
        if transformer.shape != () or transformer.cond_shape is not None:
            raise ValueError(
                "Currently, only unconditional transformers with shape () are supported."
            )

        constructor, transformer_init_params = get_ravelled_bijection_constructor(
            transformer
        )

        if cond_dim is None:
            self.cond_shape = None
            in_ranks = jnp.arange(dim)
        else:
            self.cond_shape = (cond_dim,)
            # we give conditioning variables rank -1 (no masking of edges to output)
            in_ranks = jnp.hstack((jnp.arange(dim), -jnp.ones(cond_dim)))

        hidden_ranks = tile_until_length(jnp.arange(dim), nn_width)

        out_ranks = jnp.repeat(jnp.arange(dim), transformer_init_params.size)

        autoregressive_mlp = AutoregressiveMLP(
            in_ranks,
            hidden_ranks,
            out_ranks,
            nn_depth,
            nn_activation,
            key=key,
        )

        # Initialise bias terms to match the provided transformer parameters
        self.autoregressive_mlp = eqx.tree_at(
            where=lambda t: t.layers[-1].linear.bias,
            pytree=autoregressive_mlp,
            replace=jnp.tile(
                transformer_init_params, dim
            ),
        )

        self.transformer_constructor = constructor
        self.shape = (dim,)
        self.cond_shape = None if cond_dim is None else (cond_dim,)

    def transform(self, x, condition=None):
        nn_input = x if condition is None else jnp.hstack((x, condition))
        transformer_params = self.autoregressive_mlp(nn_input)
        transformer = self._flat_params_to_transformer(transformer_params)
        return transformer.transform(x)

    def transform_and_log_abs_det_jacobian(self, x, condition=None):
        nn_input = x if condition is None else jnp.hstack((x, condition))
        transformer_params = self.autoregressive_mlp(nn_input)
        transformer = self._flat_params_to_transformer(transformer_params)
        return transformer.transform_and_log_abs_det_jacobian(x)

    def inverse(self, y, condition=None):
        init = (y, 0)
        fn = partial(self.inv_scan_fn, condition=condition)
        (x, _), _ = jax.lax.scan(fn, init, None, length=len(y))
        return x

    def inv_scan_fn(self, init, _, condition):
        "One 'step' in computing the inverse"
        y, rank = init
        nn_input = y if condition is None else jnp.hstack((y, condition))
        transformer_params = self.autoregressive_mlp(nn_input)
        transformer = self._flat_params_to_transformer(transformer_params)
        x = transformer.inverse(y)
        x = y.at[rank].set(x[rank])
        return (x, rank + 1), None

    def inverse_and_log_abs_det_jacobian(self, y, condition=None):
        x = self.inverse(y, condition)
        log_det = self.transform_and_log_abs_det_jacobian(x, condition)[1]
        return x, -log_det

    def _flat_params_to_transformer(
        self, params: Array
    ):  # TODO code repetition with MAF
        "Reshape to dim X params_per_dim, then vmap."
        dim = self.shape[-1]
        transformer_params = jnp.reshape(params, (dim, -1))
        transformer = eqx.filter_vmap(self.transformer_constructor)(transformer_params)
        return Vmap(transformer, (dim,))



from typing import Any, Sequence
import jax
import jax.numpy as jnp
import equinox as eqx
import jax.flatten_util as jfu

Array = Any  # TODO Can use TypeAlias from typing in future.


def _identity(x):
    return x


def tile_until_length(x: Array, max_len: int):
    """Tile array along until a maximum length is reached. If x.ndim > 1,
    then the array is flattened, and a flattened array is returned.

    Args:
        x (Array): Input array.
        max_len (int): Maximum length of output array.
    """
    x = jnp.ravel(x)
    num_reps = max_len // len(x) + 1
    y = jnp.tile(x, num_reps)
    return y[:max_len]


def real_to_increasing_on_interval(
    arr: Array, B: float = 1, softmax_adjust: float = 1e-2
):
    """Transform unconstrained parameter vector to monotonically increasing positions on [-B, B].

    Args:
        arr (Array): Parameter vector.
        B (float, optional): Interval to transform output. Defaults to 1.
        softmax_adjust (float, optional): Rescales softmax output using (widths + softmax_adjust/widths.size) / (1 + softmax_adjust). e.g. 0=no adjustment, 1=average softmax output with evenly spaced widths, >1 promotes more evenly spaced widths.
    """
    if softmax_adjust < 0:
        raise ValueError("softmax_adjust should be >= 0.")
    widths = jax.nn.softmax(arr)
    widths = (widths + softmax_adjust / widths.size) / (1 + softmax_adjust)
    widths = widths.at[0].set(widths[0] / 2)
    return 2 * B * jnp.cumsum(widths) - B


def inv_cum_sum(x):
    return x - jnp.pad(x[:-1], (1, 0))


def merge_shapes(shapes: Sequence):
    """ "Broadcast shapes used in bijections and distributions.

    Namely:
        - A shape None is used to mean any shape (either unknown, unimportant, or compatible with any shape).
        - We require all shapes that are not None to have the same length, and matching values.
    """  # TODO update these docs.
    if len(shapes) == 0:
        raise ValueError("No shapes have been provided.")
    elif all(s is None for s in shapes):
        return None
    else:
        shapes = [s for s in shapes if s is not None]
        if all(s == shapes[0] for s in shapes):
            return shapes[0]
        else:
            raise ValueError("The shapes do not match.")


def check_shapes(shapes):  # TODO test this?
    shapes = [s for s in shapes if s is not None]
    if not all(s == shapes[0] for s in shapes):
        raise ValueError("The shapes do not match.")


def _get_ufunc_signature(in_shapes, out_shapes):
    """Convert a sequence of in_shapes, and out_shapes to a universal function signature.
    
    Example:
        >>> _get_ufunc_signature([(3,),(2,3)], [()])
        "(3),(2,3)->()"
    """
    def _shapes_to_str(shapes):
        result = [str(s) if len(s)!=1 else str(s).replace(",", "") for s in shapes]
        return ",".join(result).replace(" ", "")

    in_shapes_str = _shapes_to_str(in_shapes)
    out_shapes_str = _shapes_to_str(out_shapes)
    return f"{in_shapes_str}->{out_shapes_str}"




def get_ravelled_bijection_constructor(bijection, filter_spec=eqx.is_inexact_array):
    """Given a bijection, returns a tuple containing
    1) a constructor for the bijection from a flattened array; 2) the current flattened
    parameters.

    Args:
        bijection (Bijection): Bijection.
        filter_spec: Filter function. Defaults to eqx.is_inexact_array.
    """

    params, static = eqx.partition(bijection, filter_spec)
    bias, unravel = jfu.ravel_pytree(params)

    def f(ravelled_params: Array):
        ravelled_params = ravelled_params + bias
        params = unravel(ravelled_params)
        return eqx.combine(params, static)

    return f, bias



from typing import Tuple, Union

import equinox as eqx
import jax.numpy as jnp
from jax.experimental import checkify

from flowjax.bijections import Bijection
from flowjax.utils import Array


class Invert(Bijection):
    bijection: Bijection

    def __init__(self, bijection: Bijection):
        """Invert a bijection, such that the transform methods become the inverse methods and vice versa.
        Note that in general, we define bijections such that the forward methods are preffered, i.e.
        faster/actually implemented. For training flows, we generally want the inverse method (used in
        density evaluation), to be faster. Hence it is often useful to use this class to achieve this aim.

        Args:
            bijection (Bijection): Bijection to "invert".
        """
        self.bijection = bijection
        self.shape = bijection.shape
        self.cond_shape = bijection.cond_shape

    def transform(self, x, condition=None):
        return self.bijection.inverse(x, condition)

    def transform_and_log_abs_det_jacobian(self, x, condition=None):
        return self.bijection.inverse_and_log_abs_det_jacobian(x, condition)

    def inverse(self, y, condition=None):
        return self.bijection.transform(y, condition)

    def inverse_and_log_abs_det_jacobian(self, y, condition=None):
        return self.bijection.transform_and_log_abs_det_jacobian(y, condition)


class Permute(Bijection):
    permutation: Array
    inverse_permutation: Array

    def __init__(self, permutation: Array):
        """Permutation transformation.

        Args:
            permutation (Array): An array with shape matching the array to transform,
                with elements 0-(array.size-1) representing the new order based on the
                flattened array (uses, C-like ordering).
        """
        checkify.check(
            (permutation.ravel().sort() == jnp.arange(permutation.size)).all(),
            "Invalid permutation array provided.",
        )
        self.shape = permutation.shape
        self.cond_shape = None

        indices = jnp.unravel_index(permutation.ravel(), permutation.shape)
        self.permutation = tuple(jnp.reshape(i, permutation.shape) for i in indices)
        
        inv_indices = jnp.unravel_index(jnp.argsort(permutation.ravel()), permutation.shape)
        self.inverse_permutation = tuple(jnp.reshape(i, permutation.shape) for i in inv_indices)

    def transform(self, x, condition=None):
        self._argcheck(x)
        return x[self.permutation]

    def transform_and_log_abs_det_jacobian(self, x, condition=None):
        return x[self.permutation], jnp.array(0)

    def inverse(self, y, condition=None):
        self._argcheck(y)
        return y[self.inverse_permutation]

    def inverse_and_log_abs_det_jacobian(self, y, condition=None):
        return y[self.inverse_permutation], jnp.array(0)


class Flip(Bijection):
    """Flip the input array. Condition argument is ignored."""

    def __init__(self) -> None:
        self.shape = None
        self.cond_shape = None    

    def transform(self, x, condition=None):
        return jnp.flip(x)

    def transform_and_log_abs_det_jacobian(self, x, condition=None):
        return jnp.flip(x), jnp.array(0)

    def inverse(self, y, condition=None):
        return jnp.flip(y)

    def inverse_and_log_abs_det_jacobian(self, y, condition=None):
        return jnp.flip(y), jnp.array(0)


class Partial(Bijection):
    """Applies bijection to specific indices of an input."""

    bijection: Array
    idxs: Union[int, slice, Array, tuple]

    def __init__(self, bijection: Bijection, idxs):
        """
        Args:
            bijection (Bijection): Bijection that is compatible with the subset of x indexed by idxs.
            idxs: Indices (Integer, a slice, or an ndarray with integer/bool dtype) of the transformed portion.
        """
        self.bijection = bijection
        self.idxs = idxs
        self.shape = None
        self.cond_shape = bijection.cond_shape

    def transform(self, x: Array, condition=None):
        y = self.bijection.transform(x[self.idxs], condition)
        return x.at[self.idxs].set(y)

    def transform_and_log_abs_det_jacobian(self, x: Array, condition=None):
        y, log_det = self.bijection.transform_and_log_abs_det_jacobian(
            x[self.idxs], condition
        )
        return x.at[self.idxs].set(y), log_det

    def inverse(self, y: Array, condition=None) -> Array:
        x = self.bijection.inverse(y[self.idxs], condition)
        return y.at[self.idxs].set(x)

    def inverse_and_log_abs_det_jacobian(self, y: Array, condition=None) -> Array:
        x, log_det = self.bijection.inverse_and_log_abs_det_jacobian(
            y[self.idxs], condition
        )
        return y.at[self.idxs].set(x), log_det


class EmbedCondition(Bijection):
    bijection: Bijection
    embedding_net: eqx.Module

    def __init__(
        self,
        bijection: Bijection,
        embedding_net: eqx.Module,
        raw_cond_shape: Tuple[int],
    ) -> None:
        """Use an embedding network to reduce the dimensionality of the conditioning variable.
        The returned bijection has cond_dim equal to the raw condition size.

        Args:
            bijection (Bijection): Bijection with ``bijection.cond_dim`` equal to the embedded size.
            embedding_net (eqx.Module): A callable equinox module that embeds a conditioning variable to size ``bijection.cond_dim``.
            raw_cond_shape (Union[None, Tuple[int]]): The dimension of the raw conditioning variable.
        """
        self.bijection = bijection
        self.embedding_net = embedding_net

        self.shape = bijection.shape
        self.cond_shape = raw_cond_shape

    def transform(self, x, condition=None):
        self._argcheck(x, condition)
        condition = self.embedding_net(condition)
        return self.bijection.transform(x, condition)

    def transform_and_log_abs_det_jacobian(self, x, condition=None):
        self._argcheck(x, condition)
        condition = self.embedding_net(condition)
        return self.bijection.transform_and_log_abs_det_jacobian(x, condition)

    def inverse(self, y, condition=None):
        self._argcheck(y, condition)
        condition = self.embedding_net(condition)
        return self.bijection.inverse(y, condition)

    def inverse_and_log_abs_det_jacobian(self, y, condition=None):
        self._argcheck(y, condition)
        condition = self.embedding_net(condition)
        return self.bijection.inverse_and_log_abs_det_jacobian(y, condition)



"""
Abstact base classes for the `Bijection` and `Bijection` types. Note when implementing bijections,
by convention we try to i) implement the "transform" methods as the faster/more intuitive approach 
(compared to the inverse methods); and ii) implement only the forward methods if an inverse
is not available. The `Invert` bijection can be used to invert the orientation if a fast inverse is
desired (e.g. maximum likelihood fitting of flows).
"""

from abc import ABC, abstractmethod
from typing import Optional, Tuple, Union

from equinox import Module

from flowjax.utils import Array


class Bijection(ABC, Module):

    """Bijection base class. Similar to :py:class:`~flowjax.distributions.Distribution`,
    bijections have a ``shape`` and a ``cond_shape`` attribute. To allow easy composing
    of bijections, all bijections support passing of conditioning variables (even if
    ignored). Unlike distributions, the ``shape`` attribute can be None, for cases where
    the shape may be unknown, or unimportant (e.g. the
    :py:class:`~flowjax.bijections.tanh.Tanh` bijection is compatible with any shape of
    input).

    The methods of bijections do not generally support passing of additional batch
    dimensions, however, ``jax.vmap`` or ``eqx.filter_vmap`` can be used to vmap
    specific methods if desired, and a bijection can be explicitly vectorised using the
    :py:class:`~flowjax.bijections.jax_transforms.Vmap` bijection.

    Bijections are registered as Jax PyTrees (as they are equinox modules), so are
    compatible with normal jax operations.

    Implementing a bijection

        (1) Inherit from ``Bijection``.
        (2) Define the attributes ``shape`` and ``cond_shape``
        (3) Implement the abstract methods ``transform``, ``transform_and_log_abs_det_jacobian``,
            ``inverse`` and ``inverse_and_log_abs_det_jacobian``. These should act on inputs compatible
            with the shapes ``shape`` for ``x``, and ``cond_shape`` for ``condition``.

    """

    shape: Union[None, Tuple[int]]
    cond_shape: Union[None, Tuple[int]]

    @abstractmethod
    def transform(self, x: Array, condition: Optional[Array] = None) -> Array:
        """Apply transformation."""

    @abstractmethod
    def transform_and_log_abs_det_jacobian(
        self, x: Array, condition: Optional[Array] = None
    ) -> Tuple:
        """Apply transformation and compute log absolute value of the Jacobian determinant."""

    @abstractmethod
    def inverse(self, y: Array, condition: Optional[Array] = None) -> Array:
        """Invert the transformation."""

    @abstractmethod
    def inverse_and_log_abs_det_jacobian(
        self, y: Array, condition: Optional[Array] = None
    ) -> Array:
        """Invert the transformation and compute log absolute value of the Jacobian determinant."""

    def _argcheck(self, x, condition=None):
        "Utility argcheck that can be added to bijection methods as required."
        if x.shape != self.shape:
            raise ValueError(f"Expected x.shape {self.shape}, got {x.shape}")

        if self.cond_shape is not None:
            if condition is None:
                raise ValueError("Condition should be provided")
            elif condition.shape != self.cond_shape:
                raise ValueError(
                    f"Expected condition.shape {self.cond_shape}, got {condition.shape}"
                )
